import os
from dataclasses import dataclass
from typing import Any, Union
from urllib.parse import urljoin

import requests
from dataclasses_json import dataclass_json
from marshmallow import ValidationError

from autopack.errors import AutoPackFetchError

API_URL = os.environ.get("AUTOPACK_API_URL", "https://autopack.ai/")


@dataclass_json
@dataclass
class PackResponse:
    """Class to map Pack properties from the API. Internal use only."""

    pack_id: str
    author: str
    repo: str
    module_path: str
    description: str
    name: str
    dependencies: list[str]
    source: str
    run_args: dict[str, Any]
    init_args: dict[str, Any]

    def pack_path(self) -> str:
        return f"{self.author}_{self.repo}_{self.name}".replace("-", "_")


def get_pack_details(pack_id: str) -> Union[PackResponse, None]:
    endpoint = "/api/details"

    url = urljoin(API_URL, endpoint)
    params = {"id": pack_id}

    response = requests.get(url, params=params)
    if response.status_code == 200:
        data = response.json()

        try:
            return PackResponse(**data)
        except (ValidationError, TypeError) as e:
            message = f"Pack fetch received invalid data: {e}"
            print(message)
            raise AutoPackFetchError(message)

    elif response.status_code <= 500:
        print(f"Error: {response.status_code}")
        return None
    else:
        print(f"Error: {response.status_code}")
        error_message = f"Error: {response.status_code}"
        raise AutoPackFetchError(error_message)


def pack_search(query: str) -> list[PackResponse]:
    endpoint = "/api/search"
    url = urljoin(API_URL, endpoint)
    params = {"query": query}

    response = requests.get(url, params=params)
    if response.status_code == 200:
        data = response.json()

        try:
            return [PackResponse(*datum) for datum in data]
        except (ValidationError, TypeError) as e:
            message = f"Pack fetch received invalid data: {e}"
            print(message)
            raise AutoPackFetchError(message)

    elif response.status_code <= 500:
        print(f"Error: {response.status_code}")
        return []
    else:
        print(f"Error: {response.status_code}")
        error_message = f"Error: {response.status_code}"
        raise AutoPackFetchError(error_message)



from typing import Any, Type

from langchain.tools import BaseTool

from autopack.api import PackResponse


class Pack(BaseTool):
    """
    Class to describe the Pack that is returned from get_pack calls. Wraps the Tool and provides metadata about the
    pack.
    """

    pack_id: str
    author: str
    repo: str
    module_path: str
    description: str
    name: str
    dependencies: list[str]
    source: str
    run_args: dict[str, Any]
    init_args: dict[str, Any]
    # TODO/FIXME: This isn't guaranteed to be an actual BaseTool, by design, but typechecking complains otherwise
    tool: Type[BaseTool]

    def init_tool(self, init_args: dict[str, Any] = None):
        init_args = init_args or {}
        is_valid = validate_tool_args(self.init_args, init_args)
        # TODO: Error handling on invalid args
        if is_valid:
            self.tool(**init_args)

    def _run(self, *args, **kwargs):
        is_valid = validate_tool_args(self.run_args, kwargs)
        # TODO: Error handling on invalid args
        if is_valid:
            if hasattr(self.tool, "_run"):
                return self.tool._run(self, *args, **kwargs)
            else:
                return self.run(*args, **kwargs)

    async def _arun(self, *args, **kwargs):
        if hasattr(self.tool, "_run"):
            return await self.tool._arun(*args, **kwargs)
        else:
            return await self.arun(*args, **kwargs)

    @property
    def args(self) -> dict:
        return self.run_args

    @property
    def is_single_input(self) -> bool:
        """Hack on top of BaseTool to allow 0-arg tools"""
        keys = {k for k in self.args if k != "kwargs"}
        return len(keys) <= 1

    @classmethod
    def from_pack_data(cls, tool: BaseTool, pack_data: PackResponse):
        return cls(
            pack_id=pack_data.pack_id,
            author=pack_data.author,
            repo=pack_data.repo,
            module_path=pack_data.module_path,
            description=pack_data.description,
            name=pack_data.name,
            dependencies=pack_data.dependencies,
            source=pack_data.source,
            run_args=pack_data.run_args,
            init_args=pack_data.init_args,
            tool=tool,
        )


def validate_tool_args(spec: dict[str, Any], actual: dict[str, Any]):
    # TODO: This. Plus maybe some helpful errors?
    return True



import importlib
import inspect
import os
import sys
from types import ModuleType
from typing import Type, Union

from autopack.api import PackResponse, get_pack_details
from autopack.errors import (
    AutoPackError,
    AutoPackLoadError,
    AutoPackNotFoundError,
    AutoPackNotInstalledError,
)
from autopack.pack import Pack


def try_get_pack(pack_id: str, quiet=False) -> Union[Pack, None]:
    """
    Get a pack based on its ID, in `author/repo_name/pack_name` format. Same as `get_pack` but does not raise an
    Exception. If there is a problem finding or loading a pack it will return None.

    Args:
        pack_id (str): The ID of the pack to fetch.
        quiet (bool, Optional): If True, won't print any output

    Returns:
        BaseTool or None: The fetched pack as a LangChain BaseTool object, None if the pack could not be loaded
    """

    try:
        return get_pack(pack_id, quiet=quiet)
    except AutoPackError:
        return None


def get_pack(pack_id: str, quiet=False) -> Pack:
    """
    Get a pack based on its ID, in `author/repo_name/pack_name` format.

    Args:
        pack_id (str): The ID of the pack to fetch.
        quiet (bool, Optional): If True, won't print any output

    Returns:
        BaseTool: The fetched pack as a LangChain BaseTool object

    Raises:
        AutoPackFetchError: If there was an error during the data fetch.
        AutoPackNotInstalledError: If the pack was found but not installed.
        AutoPackNotFoundError: If no pack matching that ID was found.
        AutoPackLoadError: If the pack was found but there was an error importing or finding the pack class.
    """
    pack_data = get_pack_details(pack_id)

    if not pack_data:
        raise AutoPackNotFoundError()

    return find_pack(pack_data, quiet=quiet)


def find_module(pack_data: PackResponse) -> ModuleType:
    module_path = pack_data.module_path

    # Just in case they already have it in their path for whatever reason
    try:
        return importlib.import_module(module_path)
    except:
        autopack_dir = find_or_create_autopack_dir()
        sys.path.insert(0, autopack_dir)

        try:
            return importlib.import_module(pack_data.pack_path() + "." + module_path)
        finally:
            sys.path.remove(autopack_dir)


def find_or_create_autopack_dir(depth=0) -> str:
    """Search in current and parent directories looking for .autopack"""
    autopack_dir = os.path.abspath(
        os.path.join(os.getcwd(), *[os.pardir] * depth, ".autopack")
    )

    if not os.path.exists(autopack_dir) or not os.path.isdir(autopack_dir):
        if depth > 3:
            raise ModuleNotFoundError(".autopack directory could not be found")
        return find_or_create_autopack_dir(depth=depth + 1)

    return autopack_dir


def find_pack(pack_data: PackResponse, quiet=False) -> Pack:
    try:
        module = find_module(pack_data)
        if not module:
            message = (
                f"Pack {pack_data.pack_id} could not be found. Either it is misconfigured or .autopack directory not "
                f"found"
            )
            raise AutoPackNotFoundError(message)

        for _, obj in inspect.getmembers(module):
            if is_valid_pack(obj, pack_data.name):
                return Pack(**{"tool": obj, **pack_data.__dict__})

        message = f"Pack {pack_data.pack_id} found, but {pack_data.name} is not found in its module"
        raise AutoPackNotFoundError(message)
    except ModuleNotFoundError:
        message = f"Pack {pack_data.pack_id} is available but not installed. To install: autopack install {pack_data.pack_id}"
        if not quiet:
            print(message)
        raise AutoPackNotInstalledError(message)
    except (ImportError, AttributeError) as e:
        message = f"Error loading {pack_data.pack_id}: {e}"
        if not quiet:
            print(message)
        raise AutoPackLoadError(message)


def is_valid_pack(klass: Type, name: str):
    if not inspect.isclass(klass):
        return False

    base_class_names = [k.__name__ for k in klass.__bases__]
    roughly_adheres_to_interface = (
        hasattr(klass, "run") or "BaseTool" in base_class_names
    )
    if not roughly_adheres_to_interface:
        return False

    # Pack name is the class name
    if name == klass.__name__:
        return True

    # Pack class has a name class variable
    if hasattr(klass, "name"):
        if callable(klass.name):
            klass_name = klass.name()
        else:
            klass_name = klass.name
        return klass_name == name

    # Pack class has a __fields__ class variable (e.g. LangChain's BaseTool)
    if hasattr(klass, "__fields__"):
        name_field = klass.__fields__.get("name")
        return name_field and name_field.default == name

    return False



import argparse

from autopack.installation import install_pack


def parse_args():
    parser = argparse.ArgumentParser(description="AutoPack CLI tool")

    subparsers = parser.add_subparsers(dest="command")
    install_parser = subparsers.add_parser("install", help="Install a pack")
    install_parser.add_argument(
        "pack", help="ID (author/repo/pack_name) of the pack to install"
    )
    parser.add_argument(
        "-f",
        "--force",
        help="Force automatic dependency installation",
        action="store_true",
    )

    return parser.parse_args()


def main():
    args = parse_args()

    if args.command == "install":
        result = install_pack(args.pack, args.force)
        if result:
            print("Installation completed")
        else:
            print("Installation failed")


if __name__ == "__main__":
    main()
